package info.textgrid.services.aggregator;

import info.textgrid.middleware.confclient.ConfservClient;
import info.textgrid.middleware.confclient.ConfservClientConstants;
import info.textgrid.middleware.tgsearch.client.SearchClient;
import info.textgrid.namespaces.metadata.core._2010.MetadataContainerType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.TGCrudService;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.TGCrudService_Service;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.UriBuilder;
import javax.xml.stream.XMLStreamException;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.MTOMFeature;

import org.codehaus.jettison.json.JSONException;

/*
 * Generating a TEIcorpus document from a (single) aggregation
 * 
 * Header:
 * 
 * Header could be generated in the same way as in the metadata editor, i.e.
 * paste together the ancestor metadata records (including matching work
 * records) and then run an XSLT stylesheet.
 * 
 * Children's Header:
 * 
 * For descendant aggregations, we still need a header. Either we consider this
 * redundant, or we preserve the metadata records we copied together for the
 * respective ancestor levels, append the current level & run the
 * transformation.
 * 
 * Body:
 * 
 * Naïve approach: Load and recursively process the aggregations, check each
 * one's content type Probably better: SearchClient has a listAggregation()
 * method that is also used by the navigator and that returns metadata records.
 * We could use this, so it's only one large request per aggregation. Should be
 * reasonably fast.
 */
@Path("/teicorpus")
public class TEICorpus {

	private static final String CONF_ENDPOINT = "https://www.textgridlab.org/1.0/confserv";
	private ConfservClient confservClient;
	private SearchClient searchClient;
	private HashMap<String, String> config;
	private TGCrudService crud;
	final Logger logger = Logger.getLogger("info.textgrid.services.aggregator");

	@GET
	@Path(value="/{aggregation}")
	@Produces("text/xml")
	public Response get(@PathParam("aggregation") final URI uri, @QueryParam("attach") @DefaultValue("true") final boolean attach) throws URISyntaxException {
		logger.fine("TEIcorpus called for root aggregation: " + uri);
		final TGCrudService crud = getCRUDService();
		logger.finest("Yo, clients are there.");
		try {
			final MetadataContainerType rootAggregationMetadata = crud.readMetadata(null, null, uri.toString());
			logger.finer("CRUD request for root aggregation successful");
			final TEICorpusSerializer serializer = new TEICorpusSerializer(rootAggregationMetadata.getObject(), this);
			final String format = rootAggregationMetadata.getObject().getGeneric().getProvided().getFormat();
			if (!format.contains("aggregation")) {
				logger.log(Level.SEVERE, "The requested object {0} is a {1}, not an aggregation, and thus cannot be aggregated.", new Object[] {uri,format});
				throw new WebApplicationException(javax.ws.rs.core.Response
						.status(Status.BAD_REQUEST)
						.entity(MessageFormat.format(
								"The requested object {0} is a {1}, not an aggregation, and thus cannot be aggregated.", uri,
								format)).build());
			}
			ResponseBuilder builder = Response.ok(serializer, "text/xml");
			if (attach) {
				String fileName = rootAggregationMetadata.getObject().getGeneric().getProvided().getTitle().get(0) + ".xml";
				String asciiFileName = fileName.replaceAll("[^A-Za-z0-9.-]", "_");
					String extendedNameSpec = "UTF-8''" + UriBuilder.fromPath(fileName).build().toASCIIString();
					builder.header("Content-Disposition", "attachment; filename*=" + extendedNameSpec + " ");
				builder.header("Content-Disposition", " filename=\"" + asciiFileName + "\" " );
			}
			return builder.build();
		} catch (final ObjectNotFoundFault e) {
			logger.log(Level.SEVERE, "CRUD: (Root aggregation) not found", e);
			throw new WebApplicationException(e, javax.ws.rs.core.Response.status(Status.NOT_FOUND).entity(e.toString()).build());
		} catch (final MetadataParseFault e) {
			logger.log(Level.SEVERE, "CRUD: Could not parse metadata", e);
			throw new WebApplicationException(e);
		} catch (final IoFault e) {
			logger.log(Level.SEVERE, "CRUD: IO Fault", e);
			throw new WebApplicationException(e);
		} catch (final AuthFault e) {
			logger.log(Level.SEVERE, "CRUD: (Root aggregation) AuthFault", e);
			throw new WebApplicationException(e, Status.FORBIDDEN);
		}

		// final Response response = search.listAggregation(uri.toString());
		// for (final ResultType result : response.getResult()) {
		// final String contentType =
		// result.getObject().getGeneric().getProvided().getFormat();
		// if (contentType.contains("aggregation")) {
		// // TODO deal with aggregation
		// } else if ("text/xml".equals(contentType)) {
		// // TODO deal with
		// }
		// }
		// return null;
	}

	TGCrudService getCRUDService() {
		if (crud == null) {
			final URL wsdl = TGCrudService_Service.class.getResource("/wsdl/TGCrudService.wsdl");
			final TGCrudService_Service tgCrudService_Service = new TGCrudService_Service(wsdl);
			final TGCrudService _crud = tgCrudService_Service.getTGCrudPort(new MTOMFeature());
			final BindingProvider bp = (BindingProvider) _crud;
			final Map<String, Object> requestContext = bp.getRequestContext();
			requestContext.put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, getConfValue(ConfservClientConstants.TG_CRUD));
			crud = _crud;
		}
		return crud;
	}

	SearchClient getSearchClient() {
		if (searchClient == null)
			searchClient = new SearchClient(getConfValue(ConfservClientConstants.TG_SEARCH_PUBLIC));
		return searchClient;
	}

	String getConfValue(final String key) throws WebApplicationException {
		if (config == null) {
			try {
				confservClient = new ConfservClient(CONF_ENDPOINT);
				config = confservClient.getAll();
			} catch (final IOException e) {
				throw new WebApplicationException(e);
			} catch (final JSONException e) {
				throw new WebApplicationException(e);
			} catch (final XMLStreamException e) {
				throw new WebApplicationException(e);
			}
		}
		return config.get(key);
	}

}
